
#include <windows.h>

#include <nkintr.h>
//#include <oalintr.h>
#include <pm.h>
#include "pmplatform.h"
#include <ceddk.h>

#include <s3c2440a.h>
#include <bsp.h>
#include "camif.h"
#include "camera.h"

#if (CAM_USE_IO == USE_SCCB_IO)
	#include "sccb.cpp"
#endif

#if	(CIS_TYPE	==	CIS_SET100B)
	#include "set100b.h"
#else
	#include "s5x532.h"
#endif	

DWORD preview_frame_count, codec_frame_count;

#ifdef DEBUG
DBGPARAM dpCurSettings = {
    TEXT("CAMERA"), {
        TEXT("0"),TEXT("1"),TEXT("2"),TEXT("3"),
        TEXT("4"),TEXT("5"),TEXT("6"),TEXT("7"),
        TEXT("8"),TEXT("9"),TEXT("10"),TEXT("11"),
        TEXT("12"),TEXT("Function"),TEXT("Init"),TEXT("Error")},
    0x8000  // Errors only, by default
}; 
#endif


static void Delay(USHORT count)
{
	volatile int i, j = 0;
	volatile static int loop = S3C2440A_FCLK/100000;
	
	for(;count > 0;count--)
		for(i=0;i < loop; i++) { j++; }
}


// 
// Read PCF50606 registers directly into our cache
//
DWORD
HW_ReadRegisters(
    PCIS_CONTEXT pCIS,
    PUCHAR pBuff,       // Optional buffer
    UCHAR StartReg,     // Start Register
    DWORD nRegs         // Number of Registers
    )
{
    DWORD dwErr;


//    EnterCriticalSection(&pCIS->RegCS);

#if (CAM_USE_IO == USE_SCCB_IO)
    if (!pBuff) {
        dwErr = ERROR_INVALID_PARAMETER;
    } else {
        dwErr = SCCB_Read(CAMERA_READ, StartReg, pBuff, nRegs);
    }
#else
    // use the driver-to-driver call
    dwErr = pCIS->fc.I2CRead(pCIS->fc.Context,
                             CAMERA_READ, 	// SlaveAddress
                             StartReg,      // WordAddress
                             !pBuff ? (PUCHAR)&pCIS->Reg[StartReg] : pBuff,
                             nRegs);
#endif    
    
    if ( !dwErr ) {


    } else {        
        RETAILMSG(1,(TEXT("I2CRead ERROR: %u \r\n"), dwErr));
        //DEBUGMSG(ZONE_ERR,(TEXT("I2CRead ERROR: %u \r\n"), dwErr));
    }            

//    LeaveCriticalSection(&pCIS->RegCS);
    
    return dwErr;
}


// 
// Write PCF50606 registers directly from our cache
//
DWORD
HW_WriteRegisters(
    PCIS_CONTEXT pCIS,
    PUCHAR pBuff,   // Optional buffer
    UCHAR StartReg, // start register
    DWORD nRegs     // number of registers
    )
{
    DWORD dwErr;


//    EnterCriticalSection(&pCIS->RegCS);

#if (CAM_USE_IO == USE_SCCB_IO)
    if (!pBuff) {
        dwErr = ERROR_INVALID_PARAMETER;
    } else {
        dwErr = SCCB_Write(CAMERA_WRITE, StartReg, pBuff, nRegs);
    }
#else
    // use the driver-to-driver call
    dwErr = pCIS->fc.I2CWrite(pCIS->fc.Context,
                              CAMERA_WRITE,   	// SlaveAddress
                              StartReg,         // WordAddress
                              !pBuff ? (PUCHAR)&pCIS->Reg[StartReg] : pBuff,
                              nRegs);
#endif

    if ( dwErr ) {
        RETAILMSG(1, (TEXT("I2CWrite ERROR: %u \r\n"), dwErr));
        //DEBUGMSG(ZONE_ERR, (TEXT("I2CWrite ERROR: %u \r\n"), dwErr));
    }            

//    LeaveCriticalSection(&pCIS->RegCS);
    
    return dwErr;
}

void CAM_WriteBlock(void)
{
    int i = 0;
    //unsigned char rdata[256] = { 0, }; 
    //int err = 0;
#if (CIS_TYPE == CIS_S5X3A1)
	for(i=0; i<(sizeof(S5X3A1_YCbCr8bit)/2); i++)
	{
		//HW_WriteRegisters(&S5X3A1_YCbCr8bit[i][1], S5X3A1_YCbCr8bit[i][0], 1);
		HW_WriteRegisters(pCIS, &S5X3A1_YCbCr8bit[i][1], S5X3A1_YCbCr8bit[i][0], 1);
	}
#elif (CIS_TYPE == CIS_S5X3AA) 
	for(i=0; i<(sizeof(S5X3AA_YCbCr8bit)/2); i++)
	{
		//HW_WriteRegisters(&S5X3AA_YCbCr8bit[i][1], S5X3AA_YCbCr8bit[i][0], 1);
		HW_WriteRegisters(pCIS, &S5X3AA_YCbCr8bit[i][1], S5X3AA_YCbCr8bit[i][0], 1);
	}
#elif (CIS_TYPE == CIS_OV9650) 
	i = 0;
	HW_ReadRegisters(pCIS, (PUCHAR)&i, 0xB, 1);
	HW_ReadRegisters(pCIS, (PUCHAR)&i+1, 0xA, 1);
	RETAILMSG(1,(TEXT("Camera::Product id is 0x%x\n"),i));

	for(i=0; i<(sizeof(OV9650Reg)/2); i++)
	{
		if (OV9650Reg[i][0] == CHIP_DELAY) Delay(OV9650Reg[i][1]);
		else HW_WriteRegisters(pCIS, &OV9650Reg[i][1], OV9650Reg[i][0], 1);
	}
	i = 0;
	HW_ReadRegisters(pCIS, (PUCHAR)&i, 0xB, 1);
	HW_ReadRegisters(pCIS, (PUCHAR)&i+1, 0xA, 1);
	RETAILMSG(1,(TEXT("Camera::Product id is 0x%x\n"),i));
#elif (CIS_TYPE == CIS_OV7660) 
	i = 0;
	HW_ReadRegisters(pCIS, (PUCHAR)&i, 0xB, 1);
	HW_ReadRegisters(pCIS, (PUCHAR)&i+1, 0xA, 1);
	RETAILMSG(1,(TEXT("Camera::Product id is 0x%x\n"),i));

	for(i=0; i<(sizeof(OV7660Reg)/2); i++)
	{
		if (OV7660Reg[i][0] == CHIP_DELAY) Delay(OV7660Reg[i][1]);
		else HW_WriteRegisters(pCIS, &OV7660Reg[i][1], OV7660Reg[i][0], 1);
	}
	i = 0;
	HW_ReadRegisters(pCIS, (PUCHAR)&i, 0xB, 1);
	HW_ReadRegisters(pCIS, (PUCHAR)&i+1, 0xA, 1);
	RETAILMSG(1,(TEXT("Camera::Product id is 0x%x\n"),i));
#elif (CIS_TYPE == CIS_SET100B)
	HW_ReadRegisters(pCIS, (PUCHAR)&i, 0x01, 1);
	HW_ReadRegisters(pCIS, (PUCHAR)&i+1, 0x02, 1);
	RETAILMSG(1,(TEXT("Camera::Product id is 0x%x.(SET100B)\n"),i));
	SIV100B_Write_Sensor_Initial_Setting();
#endif
 	/*
    for(i=0; i<(sizeof(S5X532_YCbCr8bit_TV)/2); i++){
    	RETAILMSG(I2C_MSG,(TEXT("num:%d "),i));	
		//Wr_CamIIC(SlaveID, S5X532_YCbCr8bit_TV[i][0], S5X532_YCbCr8bit_TV[i][1]);
		HW_WriteRegisters(pCIS, &S5X532_YCbCr8bit_TV[i][1], S5X532_YCbCr8bit_TV[i][0], 1);
		//Sleep(1);
	}
	*/
#if 0
	i = 0;
	HW_WriteRegisters(pCIS, &S5X532_YCbCr8bit_TV[i][1], S5X532_YCbCr8bit_TV[i][0], 1);
	do 
	{
		i++;
		HW_ReadRegisters(pCIS, &rdata[i], S5X532_YCbCr8bit_TV[i][0], 1);
		RETAILMSG(1,(_T("rdata[%d] = 0x%X\r\n"), i, rdata[i]));
		Sleep(1);
	} while (i < 10);

#endif
	
    RETAILMSG(1,(TEXT("Camera::Block TX Ended...\r\n")));	
}

/********************************************************
 CalculateBurstSize - Calculate the busrt lengths
 
 Description:	
 - dstHSize: the number of the byte of H Size.
 
*/
void CalculateBurstSize(unsigned int hSize,unsigned int *mainBurstSize,unsigned int *remainedBurstSize)
{
	unsigned int tmp;	
	tmp=(hSize/4)%16;
	switch(tmp) {
		case 0:
			*mainBurstSize=16;
			*remainedBurstSize=16;
			break;
		case 4:
			*mainBurstSize=16;
			*remainedBurstSize=4;
			break;
		case 8:
			*mainBurstSize=16;
			*remainedBurstSize=8;
			break;
		default: 
			tmp=(hSize/4)%8;
			switch(tmp) {
				case 0:
					*mainBurstSize=8;
					*remainedBurstSize=8;
					break;
				case 4:
					*mainBurstSize=8;
					*remainedBurstSize=4;
				default:
					*mainBurstSize=4;
					tmp=(hSize/4)%4;
					*remainedBurstSize= (tmp) ? tmp: 4;
					break;
			}
			break;
	}		    	    		
}

/********************************************************
 CalculatePrescalerRatioShift - none
 
 Description:	
 - none
 
*/
void CalculatePrescalerRatioShift(unsigned int SrcSize, unsigned int DstSize, unsigned int *ratio,unsigned int *shift)
{
	if(SrcSize>=64*DstSize) {
//		Uart_Printf("ERROR: out of the prescaler range: SrcSize/DstSize = %d(< 64)\r\n",SrcSize/DstSize);
		while(1);
	}
	else if(SrcSize>=32*DstSize) {
		*ratio=32;
		*shift=5;
	}
	else if(SrcSize>=16*DstSize) {
		*ratio=16;
		*shift=4;
	}
	else if(SrcSize>=8*DstSize) {
		*ratio=8;
		*shift=3;
	}
	else if(SrcSize>=4*DstSize) {
		*ratio=4;
		*shift=2;
	}
	else if(SrcSize>=2*DstSize) {
		*ratio=2;
		*shift=1;
	}
	else {
		*ratio=1;
		*shift=0;
	}    	
}




void Camif_Capture(int cap_a, int cap_b)
{
	//RETAILMSG(1,(_T("Camif_Capture(%d, %d)\r\n"), cap_a, cap_b));
	// S3C2440A

	s2440CAM->CIIMGCPT &= ~(CAM_CAMIF_GLOBAL_CAPTURE_ENABLE_BIT);

	if(cap_b == CAPTURE_ON) 
	{
		//Codec capture start
		s2440CAM->CICOSCCTRL |=(CAM_CODEC_SACLER_START_BIT);
		s2440CAM->CIIMGCPT |=(CAM_CAMIF_GLOBAL_CAPTURE_ENABLE_BIT)|(CAM_CODEC_SCALER_CAPTURE_ENABLE_BIT);
	} 
	else if (cap_b == CAPTURE_OFF)
	{
		s2440CAM->CICOSCCTRL &= ~(CAM_CODEC_SACLER_START_BIT);
		s2440CAM->CIIMGCPT &= ~(CAM_CODEC_SCALER_CAPTURE_ENABLE_BIT);
	}

	if(cap_a == CAPTURE_ON) 
	{  
		// Preview capture Start
		s2440CAM->CIPRSCCTRL |=(CAM_PVIEW_SACLER_START_BIT);
		s2440CAM->CIIMGCPT |=(CAM_CAMIF_GLOBAL_CAPTURE_ENABLE_BIT)|(CAM_PVIEW_SCALER_CAPTURE_ENABLE_BIT);
	}
	else if (cap_a == CAPTURE_OFF)
	{
		s2440CAM->CIPRSCCTRL &= ~(CAM_PVIEW_SACLER_START_BIT);
		s2440CAM->CIIMGCPT &= ~(CAM_PVIEW_SCALER_CAPTURE_ENABLE_BIT);
	}
	
	if ((cap_a == CAPTURE_ON) || (cap_b == CAPTURE_ON))
	{
		s2440CAM->CIIMGCPT |= (CAM_CAMIF_GLOBAL_CAPTURE_ENABLE_BIT);
	}
	
}



void CamIdle(void)
{
	RETAILMSG(1, (_T("Camra into idle...\r\n")));
#if 0
	if (pCIS->State != RUN)
		return;

	pCIS->m_dwTimeout = INFINITE;			// reset timeout until Camera Interrupt occurs

	// Camera clock off
	CamClockOff();

	// GPIO control
	// for power down mode, gpio goes to input mode
	s2440IOP->GPJCON = (s2440IOP->GPJCON & ~(0xffffff)) | 0x555555; // CAM IO setup except CAMRESET
	s2440IOP->GPJDAT = s2440IOP->GPJDAT & ~(0xfff); 	// CAM IO DATA pin setup except CAMRESET
	s2440IOP->GPJUP = (s2440IOP->GPJUP | 0xfff); 		// CAM IO PullUp setup except CAMRESET
	//s2440IOP->GPJCON = (s2440IOP->GPJCON & ~(0x3<<22)) | (0x0<<22); // CAMCLKOUT -> input
#endif	
}

void CamExitIdle(void)
{
	if (pCIS->State != IDLE)
		return;

	RETAILMSG(1, (_T("Camra exit idle...\r\n")));
#if 0	
	// GPIO control
	s2440IOP->GPJCON = (s2440IOP->GPJCON & ~(0xffffff)) | 0xaaaaaa; // CAM IO setup except CAMRESET
	s2440IOP->GPJDAT = s2440IOP->GPJDAT & ~(0xfff); 	// CAM IO DATA pin setup except CAMRESET
	s2440IOP->GPJUP = (s2440IOP->GPJUP | 0xfff); 		// CAM IO PullUp setup except CAMRESET
	//s2440IOP->GPJCON = (s2440IOP->GPJCON & ~(0x3<<22)) | (0x2<<22); // CAMCLKOUT -> input

	Delay(10);

	// Camera clock on
	CamSetClock(CAM_CLK_DIV);
	CamClockOn();
	
	Delay(1000);
#endif	

}


DWORD CameraCaptureThread(void)
{
	unsigned char tmp=0;
	static unsigned int time,old_time;
	static unsigned int cam_intr;
	
	DWORD	dwCause;					// LJY PWR040613

	SetProcPermissions((DWORD)-1);
	
	while(TRUE)
	{
		RETAILMSG(0,(TEXT("[CAM_HW] InterruptThread : Waiting For a Single Object\n\r")));
		
		dwCause = WaitForSingleObject(pCIS->m_hEvent, pCIS->m_dwTimeout);
		
#if 0
		if (preview_frame_count <= 2) {
			//preview_frame_count++;
			// Enable camera interrupt
			s2440INT->INTSUBMSK &= ~(( 1 << IRQ_SUB_CAM_P )|( 1 << IRQ_SUB_CAM_C ));
			s2440INT->INTMSK &= ~( 1 << IRQ_CAM );
			
			continue;
		}
#endif
		
		if (dwCause == WAIT_OBJECT_0) // LJY PWR
		{

			__try{
				//if (s2440INT->INTSUBMSK & ( 1 << IRQ_SUB_CAM_C ))
				if (s2440INT->SUBSRCPND & (1<< IRQ_SUB_CAM_C))	{
					codec_frame_count++;
					cam_intr |= ( 1 << IRQ_SUB_CAM_C );
				    
				    s2440INT->SUBSRCPND  =  (1<<IRQ_SUB_CAM_C);
		            s2440INT->INTSUBMSK &= ~(1<<IRQ_SUB_CAM_C);
					//RETAILMSG(1,(_T("CAM_C, ts %d\r\n"), GetTickCount()));
				}

				//if (s2440INT->INTSUBMSK & ( 1 << IRQ_SUB_CAM_P ))
				if (s2440INT->SUBSRCPND & (1 << IRQ_SUB_CAM_P))	{
					cam_intr |= ( 1 << IRQ_SUB_CAM_P );
					preview_frame_count++;
					s2440INT->SUBSRCPND  =  (1<<IRQ_SUB_CAM_P);
		            s2440INT->INTSUBMSK &= ~(1<<IRQ_SUB_CAM_P);
					//RETAILMSG(1,(_T("CAM_P, ts %d\r\n"), GetTickCount()));
				}

				//if (((s2440INT->INTSUBMSK & ( 1 << IRQ_SUB_CAM_C )) == 0) && ((s2440INT->INTSUBMSK & ( 1 << IRQ_SUB_CAM_P )) == 0))
				//{
				//	RETAILMSG(MSG_EN_1,(_T("[CAM]NOP\r\n")));					
				//}
				if (((s2440INT->SUBSRCPND & ( 1 << IRQ_SUB_CAM_C )) == 0) && ((s2440INT->SUBSRCPND & ( 1 << IRQ_SUB_CAM_P )) == 0))
				{
					RETAILMSG(MSG_EN_1,(_T("[CAM]NOP\r\n")));					
				}
				
				
				InterruptDone(pCIS->m_dwSysintr);				
			
				//time = GetTickCount();
				//RETAILMSG(1,(TEXT("+time:%d\r\n"),(time - old_time)));
	
				// Handle any interrupts on the input source
				if (cam_intr & ( 1 << IRQ_SUB_CAM_P ))	{
					// display the image	
					//if (DRIVER_PREVIEW_ENABLE == 1)
					
					//Display_Cam_Image(0,0,sCAMINFO.nPreWidth, sCAMINFO.nPreHeight);

					Buffer_preview_info_update();
					cam_intr &= ~( 1 << IRQ_SUB_CAM_P );
				}
					
				if (cam_intr & ( 1 << IRQ_SUB_CAM_C ))	{
					Buffer_codec_info_update();
					cam_intr &= ~( 1 << IRQ_SUB_CAM_C );
				}

				// Enable camera interrupt
				//s2440INT->INTSUBMSK &= ~(( 1 << IRQ_SUB_CAM_P )|( 1 << IRQ_SUB_CAM_C ));
				//s2440INT->INTMSK &= ~( 1 << IRQ_CAM );

				//old_time = GetTickCount();
				//RETAILMSG(1,(TEXT("-time:%d\r\n"),(old_time-time)));

			}	__except(EXCEPTION_EXECUTE_HANDLER) {
				RETAILMSG(PM_MSG, (TEXT("Camera.DLL:InterruptThread() - EXCEPTION: %d"), GetExceptionCode()));
			}
							
		}else if (dwCause == WAIT_TIMEOUT)		{
			RETAILMSG(1,(L"Thread Timeout : %d msec. idle...\r\n", pCIS->m_dwTimeout));

			if (TryEnterCriticalSection(&pCIS->CS)) { // without blocking.
				if (pCIS->State == RUN) {
					CamIdle();
					pCIS->State = IDLE;
				}
				LeaveCriticalSection(&pCIS->CS);
			}
			
		} else {
			RETAILMSG(PM_MSG, (TEXT("[CAM_HW] InterruptThread : Exit %d, Cause %d\r\n"), GetLastError(), dwCause));
		}
	}

	return 0;
}

void Display_Cam_Image(U32 pos_x, U32 pos_y, U32 size_x, U32 size_y)
{
	U8 *buffer_rgb;
	U32 y;
	int temp;
	
	//unsigned short *ptr = (unsigned short *)(FRAMEBUF_BASE+0x5dc0);
	
	//static unsigned short transfer_data[QCIF_XSIZE*QCIF_YSIZE];
	static unsigned int time,old_time;
	
	RETAILMSG(MSG_EN_1,(_T("Display_Cam_Image()\r\n")));
	
	temp = (s2440CAM->CIPRSTATUS>>26)&3;
	temp = (temp + 2) % 4;
		
	switch (temp)
	{
		case 0:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA1;
			break;
		case 1:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA2;
			break;
		case 2:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA3;
			break;
		case 3:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA4;
			break;
		default :
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA1;
			break;
	}

	RETAILMSG(MSG_EN_1,(_T("preview buf index = %d\r\n"), temp));

	
	buffer_rgb += VIRTUAL_OFFSET;

//	time = GetTickCount();
//	RETAILMSG(1,(TEXT("+:%d\r\n"),(time - old_time)));


#if 1
SetKMode(TRUE);
	for (y=0; y<size_y; y++) // YCbCr 4:2:0 format
	{
		//memcpy((void *)(FRAMEBUF_BASE+0x5e00+y*240*2),(void *)buffer_rgb,(QCIF_XSIZE)*2);
		memcpy((void *)(IMAGE_FRAMEBUFFER_UA_BASE + y*240*2),
						(void *)buffer_rgb,
						(size_x)*2);
		buffer_rgb += (size_x*2);
	}
SetKMode(FALSE);

#endif


//	old_time = GetTickCount();
//	RETAILMSG(1,(TEXT("-:%d\r\n"),(old_time - time)));

}
void CamCopyImage(PBYTE buf)
{
	U8 *buffer_rgb;
	int temp;
	
	temp = (s2440CAM->CIPRSTATUS>>26)&3;
	temp = (temp + 2) % 4;
		
	switch (temp)
	{
		case 0:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA1;
			break;
		case 1:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA2;
			break;
		case 2:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA3;
			break;
		case 3:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA4;
			break;
		default :
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA1;
			break;
	}

	RETAILMSG(MSG_EN_1,(_T("preview buf index = %d\r\n"), temp));

	
	buffer_rgb += VIRTUAL_OFFSET;
	SetKMode(TRUE);
	memcpy(buf, buffer_rgb, QVGA_XSIZE*QVGA_YSIZE*2);
	SetKMode(FALSE);

}

void Display_Cam_Image2(U32 size_x, U32 size_y)
{
	U8 *buffer_rgb;
	U32 x, y;
	int temp;
	
	//unsigned short *ptr = (unsigned short *)(FRAMEBUF_BASE+0x5dc0);
	
	//static unsigned short transfer_data[QCIF_XSIZE*QCIF_YSIZE];
	static unsigned int time,old_time;

	RETAILMSG(MSG_EN_1,(_T("Display_Cam_Image()\r\n")));
	
	temp = (s2440CAM->CIPRSTATUS>>26)&3;
	temp = (temp + 2) % 4;
		
	switch (temp)
	{
		case 0:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA1;
			break;
		case 1:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA2;
			break;
		case 2:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA3;
			break;
		case 3:
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA4;
			break;
		default :
			buffer_rgb = (U8 *)s2440CAM->CIPRCLRSA1;
			break;
	}
	
	RETAILMSG(MSG_EN_1,(_T("preview buf index = %d\r\n"), temp));

	RETAILMSG(MSG_EN_1,(_T("buffer_rgb = 0x%x\r\n"), buffer_rgb));

	
	buffer_rgb += VIRTUAL_OFFSET;

	RETAILMSG(MSG_EN_1,(_T("buffer_rgb = 0x%x\r\n"), buffer_rgb));

//	time = GetTickCount();
//	RETAILMSG(1,(TEXT("+:%d\r\n"),(time - old_time)));


#if 1
SetKMode(TRUE);
      for(y=0;y<240;y++)
      	{
		for (x=0; x<320; x++) // YCbCr 4:2:0 format
	{
		*(U16*)((U8*)(IMAGE_FRAMEBUFFER_UA_BASE+y*2+240*(319-x)*2))=*(U16*)((U8*)(buffer_rgb+y*320*2+x*2));
		//memcpy((U8 *)(FRAMEBUF_BASE + (y*2)+240*(319-x)*2),(U8 *)(buffer_rgb+y*320*2+x*2),2);
	}
      	}
SetKMode(FALSE);

#endif

}



BOOL InitInterruptThread()
{
	DWORD         threadID;                         // thread ID
	BOOL bSuccess;

    pCIS->m_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    
    if (!pCIS->m_hEvent) {
    	return FALSE;
    }

	bSuccess = InterruptInitialize(pCIS->m_dwSysintr, pCIS->m_hEvent, NULL, 0);
    if (!bSuccess) {		
        RETAILMSG(1,(TEXT("Fail to initialize camera interrupt event\r\n")));
        goto err;
    }    
	
    pCIS->m_hThread = CreateThread(NULL,
                                 0,
                                 (LPTHREAD_START_ROUTINE)CameraCaptureThread,
                                 0,
                                 0,
                                 &threadID);
    
    if (NULL == pCIS->m_hThread ) {
    	RETAILMSG(1,(TEXT("Create Camera Thread Fail\r\n")));
		goto err;
    }
	
	RETAILMSG(0,(_T("CAMERA.DLL::InterruptThread Initialized.\r\n")));
	
	return TRUE;


err:
	
	if (pCIS->m_hEvent)
		CloseHandle(pCIS->m_hEvent);
	
	return FALSE;
}


void CamGpioInit(void)
{
	s2440IOP->GPJCON = 0x2aaaaaa;
	s2440IOP->GPJDAT = 0;
	s2440IOP->GPJUP = 0;//0x1fff;
    s2440IOP->GPECON&= ~(0x3<<30); //camreset 1
    s2440IOP->GPECON|=(0x1<<30);
    s2440IOP->GPEDAT|=(0x1<<15);

	s2440IOP->GPECON&= ~(0x3<<28); //camreset 1
    s2440IOP->GPECON|=(0x1<<28);
    s2440IOP->GPEDAT|=(0x1<<14);

	
    s2440IOP->GPCCON&= ~(0x3<<10);  //campwdn  0
    s2440IOP->GPCCON|=(0x1<<10);
    s2440IOP->GPCUP|=(0x1<<5);

    //s2440IOP->GPCDAT|=(0x1<<5);
    s2440IOP->GPCDAT&=~(0x1<<5);
    Delay(1000);
	
    s2440IOP->GPJCON&= ~(0x3<<24); //camreset 1
    s2440IOP->GPJCON|=(0x1<<24);
    s2440IOP->GPJUP|=(0x1<<12);

    s2440IOP->GPJDAT|=(0x1<<12);
	Delay(2000);
    s2440IOP->GPJDAT&=~(0x1<<12);
	Delay(2000);
    s2440IOP->GPJDAT|=(0x1<<12);
	Delay(2000);
	
}



BOOL CameraInit(void)
{
	// 2. Camera i/f software reset
    CamSwReset();
 	//
    // 1. Camera IO setup
    //
	CamGpioInit();

    //
    // 3. Camera Clock setup
    //
	CamSetClock(CAM_CLK_DIV);
	CamClockOn();

	// 4. camera module reset
	CamReset();
	

	// 5. set register of camera module through iic 
	//camera_initialize();
	// use iic for initialization
	CAM_WriteBlock();

	// to check time
	//s2440IOP->GPGCON &= ~(0x3<<24);
	//s2440IOP->GPGCON |= (0x1<<24);		// EINT20

 	CamInit(sCAMINFO.nDestWidth, sCAMINFO.nDestHeight, sCAMINFO.nPreWidth, sCAMINFO.nPreHeight,\
 			0, 0, (U32)(pCIS->m_PhysCodecAddr.LowPart), (U32)(pCIS->m_PhysPreviewAddr.LowPart));

    //RETAILMSG(1,(TEXT("PhysPreviewAddress = %x, PhysCodecAddress= %x \r\n"),(U32)(pCIS->m_PhysPreviewAddr.LowPart),(U32)(pCIS->m_PhysCodecAddr.LowPart)));
	return TRUE;
}


// for MPEG4
void Copy_Cam_Image(U8 *pBufOut, U32 size_x, U32 size_y)
{
	U8 *buffer_y, *buffer_cb, *buffer_cr;
	U32 Y_size, C_size, P_size;
	int temp;
	U8 *pImage;
	static unsigned int frame_timestamp_prev = GetTickCount();
	static int frame_bank_prev = -1;
	YUVINFO yuvinfo;

	
	pImage = pBufOut;

	Y_size = sCAMINFO.nDestWidth * sCAMINFO.nDestHeight;
	C_size = Y_size/2;	
	P_size = Y_size + C_size*2;

	temp = (s2440CAM->CICOSTATUS>>26)&3;
	temp = (temp + 2) % 4;

	RETAILMSG(MSG_EN_1,(_T("Copy_Cam_Image %d\r\n"), temp));

	switch (temp)
	{
		case 0:
			buffer_y = (U8 *)s2440CAM->CICOYSA1;
			buffer_cb = (U8 *)s2440CAM->CICOCBSA1;
			buffer_cr = (U8 *)s2440CAM->CICOCRSA1;
			break;
		case 1:
			buffer_y = (U8 *)s2440CAM->CICOYSA2;
			buffer_cb = (U8 *)s2440CAM->CICOCBSA2;
			buffer_cr = (U8 *)s2440CAM->CICOCRSA2;
			break;
		case 2:
			buffer_y = (U8 *)s2440CAM->CICOYSA3;
			buffer_cb = (U8 *)s2440CAM->CICOCBSA3;
			buffer_cr = (U8 *)s2440CAM->CICOCRSA3;
			break;
		case 3:
			buffer_y = (U8 *)s2440CAM->CICOYSA4;
			buffer_cb = (U8 *)s2440CAM->CICOCBSA4;
			buffer_cr = (U8 *)s2440CAM->CICOCRSA4;
			break;
		default :
			buffer_y = (U8 *)s2440CAM->CICOYSA1;
			buffer_cb = (U8 *)s2440CAM->CICOCBSA1;
			buffer_cr = (U8 *)s2440CAM->CICOCRSA1;
			break;
	}

	buffer_y += VIRTUAL_ADDR_OFFSET;
	buffer_cb += VIRTUAL_ADDR_OFFSET;
	buffer_cr += VIRTUAL_ADDR_OFFSET;

	RETAILMSG(MSG_EN_1,(_T("buffer_y = 0x%x\r\n"), buffer_y));
	RETAILMSG(MSG_EN_1,(_T("buffer_cb = 0x%x\r\n"), buffer_cb));
	RETAILMSG(MSG_EN_1,(_T("buffer_cr = 0x%x\r\n"), buffer_cr));

	yuvinfo.frame_width = size_x;
	yuvinfo.frame_height = size_y;

	RETAILMSG(1, (L"Copy_Cam_Image w(%u) h(%u)\r\n", yuvinfo.frame_width, yuvinfo.frame_height));
	
	if (temp == frame_bank_prev) yuvinfo.frame_stamp = frame_timestamp_prev;
	else 	{
		yuvinfo.frame_stamp = GetTickCount();
		frame_timestamp_prev = yuvinfo.frame_stamp;
		frame_bank_prev = temp;
	}

	RETAILMSG(MSG_EN_1,(_T("pBufOut 0 offset = 0x%x\r\n"), pImage));
	memcpy(pImage, &yuvinfo, sizeof(YUVINFO));
	pImage += sizeof(YUVINFO);	
	
	RETAILMSG(MSG_EN_1,(_T("pBufOut Y = 0x%x\r\n"), pImage));
	memcpy(pImage, buffer_y, Y_size);
	pImage += Y_size;

	RETAILMSG(MSG_EN_1,(_T("pBufOut cb = 0x%x\r\n"), pImage));
	memcpy(pImage, buffer_cb, C_size);

	pImage += C_size;
	RETAILMSG(MSG_EN_1,(_T("pBufOut cr = 0x%x\r\n"), pImage));
	memcpy(pImage, buffer_cr, C_size);

//	pBufOut += (size_x*size_y)/4;

//	RETAILMSG(MSG_EN_1,(TEXT("1:buffer_y:0x%x,buffer_cb:0x%x,buffer_cr:0x%x\r\n"),buffer_y,buffer_cb,buffer_cr));

	//s2440IOP->GPFDAT |= (1<<6);	
	
	//Tick_COPY_FRAME = GetTickCount();

	//RETAILMSG(MSG_EN_1,(_T("COPY_Tick = %d\r\n"), (Tick_COPY_FRAME-Tick_GET_FRAME_CUR)));
	//RETAILMSG(MSG_EN_1,(_T("Frame_Tick = %d\r\n"), (Tick_GET_FRAME_CUR-Tick_GET_FRAME_PREV)));
	
//	RETAILMSG(1,(_T("Tick_COPY_FRAME = %d\r\n"), Tick_COPY_FRAME));
//	RETAILMSG(1,(_T("Tick_GET_FRAME_PREV = %d\r\n"), Tick_GET_FRAME_PREV));
//	RETAILMSG(1,(_T("Tick_GET_FRAME_CUR = %d\r\n"), Tick_GET_FRAME_CUR));	


}


void Buffer_codec_info_update()
{
	int temp;
	unsigned int buffer_y, buffer_cb, buffer_cr;
	
	temp = (s2440CAM->CICOSTATUS>>26)&3;
	temp = (temp + 2) % 4;

	switch (temp)
	{
	case 0:
		buffer_y = s2440CAM->CICOYSA1;
		buffer_cb = s2440CAM->CICOCBSA1;
		buffer_cr = s2440CAM->CICOCRSA1;
		break;
	case 1:
		buffer_y = s2440CAM->CICOYSA2;
		buffer_cb = s2440CAM->CICOCBSA2;
		buffer_cr = s2440CAM->CICOCRSA2;
		break;
	case 2:
		buffer_y = s2440CAM->CICOYSA3;
		buffer_cb = s2440CAM->CICOCBSA3;
		buffer_cr = s2440CAM->CICOCRSA3;
		break;
	case 3:
		buffer_y = s2440CAM->CICOYSA4;
		buffer_cb = s2440CAM->CICOCBSA4;
		buffer_cr = s2440CAM->CICOCRSA4;
		break;
	default :
		buffer_y = s2440CAM->CICOYSA1;
		buffer_cb = s2440CAM->CICOCBSA1;
		buffer_cr = s2440CAM->CICOCRSA1;
		break;
	}

	buffer_y += VIRTUAL_ADDR_OFFSET;
	buffer_cb += VIRTUAL_ADDR_OFFSET;
	buffer_cr += VIRTUAL_ADDR_OFFSET;
#if 0
	RETAILMSG(MSG_EN_1,(_T("buffer_y = 0x%x\r\n"), buffer_y));
	RETAILMSG(MSG_EN_1,(_T("buffer_cb = 0x%x\r\n"), buffer_cb));
	RETAILMSG(MSG_EN_1,(_T("buffer_cr = 0x%x\r\n"), buffer_cr));
#endif
	//if( codec_flag )	RETAILMSG(MSG_EN_1,(_T("Buffer is not read\r\n")));
	//codec_flag = 1;
	//y_address = buffer_y;
	//cb_address = buffer_cb;
	//cr_address = buffer_cr;
}


void Buffer_preview_info_update()
{
	int temp;
	unsigned int buffer_rgb;
	
	temp = (s2440CAM->CIPRSTATUS>>26)&3;
	temp = (temp + 2) % 4;

	switch (temp)
	{
	case 0:
		buffer_rgb = s2440CAM->CIPRCLRSA1;
		break;
	case 1:
		buffer_rgb = s2440CAM->CIPRCLRSA2;
		break;
	case 2:
		buffer_rgb = s2440CAM->CIPRCLRSA3;
		break;
	case 3:
		buffer_rgb = s2440CAM->CIPRCLRSA4;
		break;
	default :
		buffer_rgb = s2440CAM->CIPRCLRSA1;
		break;
	}

	RETAILMSG(MSG_EN_1,(_T("buffer_rgb[PHY] = 0x%x\r\n"), buffer_rgb));
	buffer_rgb += VIRTUAL_ADDR_OFFSET;
#if 1
	RETAILMSG(MSG_EN_1,(_T("buffer_rgb = 0x%x\r\n"), buffer_rgb));
#endif
	//if( rgb_flag )	RETAILMSG(MSG_EN_1,(_T("Prev Buffer is not read\r\n")));
	//rgb_flag = 1;
	//rgb_address = buffer_rgb;
}


void CamInit(U32 CoDstWidth, U32 CoDstHeight, U32 PrDstWidth, U32 PrDstHeight, 
			U32 WinHorOffset, U32 WinVerOffset,  U32 CoFrameBuffer, U32 PrFrameBuffer)
{
	U32 WinOfsEn;
	U32 divisor, multiplier;
	U32 MainBurstSizeY, RemainedBurstSizeY, MainBurstSizeC, RemainedBurstSizeC, MainBurstSizeRGB, RemainedBurstSizeRGB;
	U32 H_Shift, V_Shift, PreHorRatio, PreVerRatio, MainHorRatio, MainVerRatio;
	U32 SrcWidth, SrcHeight;
	U32 ScaleUp_H_Co, ScaleUp_V_Co, ScaleUp_H_Pr, ScaleUp_V_Pr;

	RETAILMSG(1, (L"CamInit::C-path(%d, %d),p-path(%d, %d) winoffset(%d, %d)\r\n", \
					CoDstWidth, CoDstHeight, PrDstWidth, PrDstHeight, WinHorOffset, WinVerOffset));

	//constant for calculating codec dma address
	if(CAM_CODEC_OUTPUT)
		divisor=2; //CCIR-422
	else
		divisor=4; //CCIR-420
		
	//constant for calculating preview dma address
	if(CAM_PVIEW_OUTPUT)
		multiplier=4;
	else
		multiplier=2;
	
	if(WinHorOffset==0 && WinVerOffset==0)
		WinOfsEn=0;
	else
		WinOfsEn=1;

	SrcWidth=CAM_SRC_HSIZE-WinHorOffset*2;
	SrcHeight=CAM_SRC_VSIZE-WinVerOffset*2;

	if(SrcWidth>=CoDstWidth) ScaleUp_H_Co=0; //down
	else ScaleUp_H_Co=1;		//up

	if(SrcHeight>=CoDstHeight) ScaleUp_V_Co=0;
	else ScaleUp_V_Co=1;		

	if(SrcWidth>=PrDstWidth) ScaleUp_H_Pr=0; //down
	else ScaleUp_H_Pr=1;		//up

	if(SrcHeight>=PrDstHeight) ScaleUp_V_Pr=0;
	else ScaleUp_V_Pr=1;		

	////////////////// common control setting
	s2440CAM->CIGCTRL |= (1<<26)|(0<<27); // inverse PCLK, test pattern
	s2440CAM->CIWDOFST = (1<<30)|(0xf<<12); // clear overflow 
	s2440CAM->CIWDOFST = 0;	
	s2440CAM->CIWDOFST=(WinOfsEn<<31)|(WinHorOffset<<16)|(WinVerOffset);

#if(CIS_TYPE == CIS_S5X3A1)
	s2440CAM->CISRCFMT=(CAM_ITU601<<31)|(0<<30)|(0<<29)|(CAM_SRC_HSIZE<<16)|(CAM_ORDER_YCBYCR<<14)|(CAM_SRC_VSIZE);
#elif(CIS_TYPE == CIS_S5X3AA)
    s2440CAM->CISRCFMT=(CAM_ITU601<<31)|(0<<30)|(0<<29)|(CAM_SRC_HSIZE<<16)|(CAM_ORDER_CBYCRY<<14)|(CAM_SRC_VSIZE); 
#elif(CIS_TYPE == CIS_OV7660)
	s2440CAM->CISRCFMT=(CAM_ITU601<<31)|(0<<30)|(0<<29)|(CAM_SRC_HSIZE<<16)|(CAM_ORDER_CBYCRY<<14)|(CAM_SRC_VSIZE);
#elif(CIS_TYPE == CIS_OV9650)
	s2440CAM->CISRCFMT=(CAM_ITU601<<31)|(0<<30)|(0<<29)|(CAM_SRC_HSIZE<<16)|(CAM_ORDER_YCBYCR<<14)|(CAM_SRC_VSIZE);
#elif(CIS_TYPE == CIS_SET100B)
	s2440CAM->CISRCFMT=(CAM_ITU601<<31)|(0<<30)|(0<<29)|(CAM_SRC_HSIZE<<16)|(CAM_ORDER_CBYCRY<<14)|(CAM_SRC_VSIZE);
#endif


	////////////////// codec port setting
	s2440CAM->CICOYSA1=CoFrameBuffer;
	s2440CAM->CICOYSA2=s2440CAM->CICOYSA1+CoDstWidth*CoDstHeight+2*CoDstWidth*CoDstHeight/divisor;
	if(CoDstWidth*CoDstHeight < 655360)   // Codec Max. Size = 1280*1024   Above sXGA, use only 2 memory of the 4 pingpong memory
	{
		s2440CAM->CICOYSA3=s2440CAM->CICOYSA2+CoDstWidth*CoDstHeight+2*CoDstWidth*CoDstHeight/divisor;
		s2440CAM->CICOYSA4=s2440CAM->CICOYSA3+CoDstWidth*CoDstHeight+2*CoDstWidth*CoDstHeight/divisor;
	}
	else
	{
		s2440CAM->CICOYSA3=s2440CAM->CICOYSA1;
		s2440CAM->CICOYSA4=s2440CAM->CICOYSA2;		
	}
	
	s2440CAM->CICOCBSA1=s2440CAM->CICOYSA1+CoDstWidth*CoDstHeight;
	s2440CAM->CICOCBSA2=s2440CAM->CICOYSA2+CoDstWidth*CoDstHeight;
	s2440CAM->CICOCBSA3=s2440CAM->CICOYSA3+CoDstWidth*CoDstHeight;
	s2440CAM->CICOCBSA4=s2440CAM->CICOYSA4+CoDstWidth*CoDstHeight;

	s2440CAM->CICOCRSA1=s2440CAM->CICOCBSA1+CoDstWidth*CoDstHeight/divisor;
	s2440CAM->CICOCRSA2=s2440CAM->CICOCBSA2+CoDstWidth*CoDstHeight/divisor;
	s2440CAM->CICOCRSA3=s2440CAM->CICOCBSA3+CoDstWidth*CoDstHeight/divisor;
	s2440CAM->CICOCRSA4=s2440CAM->CICOCBSA4+CoDstWidth*CoDstHeight/divisor;
    
	//s2440CAM->CICOTRGFMT=(CAM_CODEC_IN_422<<31)|(CAM_CODEC_OUTPUT<<30)|(CoDstWidth<<16)|(CAM_FLIP_NORMAL<<14)|(CoDstHeight);
	s2440CAM->CICOTRGFMT=(CAM_CODEC_IN_422<<31)|(CAM_CCIR422<<30)|(CoDstWidth<<16)|(CAM_FLIP_NORMAL<<14)|(CoDstHeight);

	CalculateBurstSize(CoDstWidth, &MainBurstSizeY, &RemainedBurstSizeY);
	CalculateBurstSize(CoDstWidth/2, &MainBurstSizeC, &RemainedBurstSizeC);
	s2440CAM->CICOCTRL=(MainBurstSizeY<<19)|(RemainedBurstSizeY<<14)|(MainBurstSizeC<<9)|(RemainedBurstSizeC<<4);

	CalculatePrescalerRatioShift(SrcWidth, CoDstWidth, &PreHorRatio, &H_Shift);
	CalculatePrescalerRatioShift(SrcHeight, CoDstHeight, &PreVerRatio, &V_Shift);
	MainHorRatio=(SrcWidth<<8)/(CoDstWidth<<H_Shift);
	MainVerRatio=(SrcHeight<<8)/(CoDstHeight<<V_Shift);
    			
	s2440CAM->CICOSCPRERATIO=((10-H_Shift-V_Shift)<<28)|(PreHorRatio<<16)|(PreVerRatio);
	s2440CAM->CICOSCPREDST=((SrcWidth/PreHorRatio)<<16)|(SrcHeight/PreVerRatio); 
	s2440CAM->CICOSCCTRL=(CAM_SCALER_BYPASS_OFF<<31)|(ScaleUp_H_Co<<30)|(ScaleUp_V_Co<<29)|(MainHorRatio<<16)|(MainVerRatio);

	s2440CAM->CICOTAREA=CoDstWidth*CoDstHeight;

	///////////////// preview port setting
	s2440CAM->CIPRCLRSA1=PrFrameBuffer;
	s2440CAM->CIPRCLRSA2=s2440CAM->CIPRCLRSA1+PrDstWidth*PrDstHeight*multiplier;
	s2440CAM->CIPRCLRSA3=s2440CAM->CIPRCLRSA2+PrDstWidth*PrDstHeight*multiplier;
	s2440CAM->CIPRCLRSA4=s2440CAM->CIPRCLRSA3+PrDstWidth*PrDstHeight*multiplier;
/*
	if(CIS_TYPE == CIS_S5X3A1)
		s2440CAM->CIPRTRGFMT=(PrDstWidth<<16)|(CAM_FLIP_YAXIS<<14)|(PrDstHeight);
	else
		s2440CAM->CIPRTRGFMT=(PrDstWidth<<16)|(CAM_FLIP_NORMAL<<14)|(PrDstHeight);*/
	//s2440CAM->CIPRTRGFMT=(PrDstWidth<<16)|(CAM_FLIP_XAXIS<<14)|(PrDstHeight);
    // target image format of preview dma.
	s2440CAM->CIPRTRGFMT=(PrDstWidth<<16)|(CAM_FLIP_NORMAL<<14)|(PrDstHeight);
	
	if (CAM_PVIEW_OUTPUT==CAM_RGB24B)
		CalculateBurstSize(PrDstWidth*4, &MainBurstSizeRGB, &RemainedBurstSizeRGB);
	else // RGB16B
		CalculateBurstSize(PrDstWidth*2, &MainBurstSizeRGB, &RemainedBurstSizeRGB);
   	s2440CAM->CIPRCTRL=(MainBurstSizeRGB<<19)|(RemainedBurstSizeRGB<<14);

	CalculatePrescalerRatioShift(SrcWidth, PrDstWidth, &PreHorRatio, &H_Shift);
	CalculatePrescalerRatioShift(SrcHeight, PrDstHeight, &PreVerRatio, &V_Shift);
	MainHorRatio=(SrcWidth<<8)/(PrDstWidth<<H_Shift);
	MainVerRatio=(SrcHeight<<8)/(PrDstHeight<<V_Shift);
	s2440CAM->CIPRSCPRERATIO=((10-H_Shift-V_Shift)<<28)|(PreHorRatio<<16)|(PreVerRatio);		 
	s2440CAM->CIPRSCPREDST=((SrcWidth/PreHorRatio)<<16)|(SrcHeight/PreVerRatio);
	s2440CAM->CIPRSCCTRL=(1<<31)|(CAM_RGB16B<<30)|(ScaleUp_H_Pr<<29)|(ScaleUp_V_Pr<<28)|(MainHorRatio<<16)|(MainVerRatio);
    
	s2440CAM->CIPRTAREA= PrDstWidth*PrDstHeight;

	// LJY, added 040615	
	// initialization for buffer addresses
	//y_address = s2440CAM->CICOYSA1;
	//cb_address = s2440CAM->CICOCBSA1;
	//cr_address = s2440CAM->CICOCRSA1;
	//rgb_address = s2440CAM->CIPRCLRSA1;
}


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
BOOL CamHandleIOControl(DWORD hOpenContext, 
				   DWORD dwCode, 
				   PBYTE pBufIn, 
				   DWORD dwLenIn, 
				   PBYTE pBufOut, 
				   DWORD dwLenOut, 
				   PDWORD pdwActualOut)
{
    BOOL RetVal = TRUE;
    DWORD dwErr = ERROR_SUCCESS;    
	//static unsigned int time=0,old_time=0;
	static int running;

	switch (dwCode)
	{

//-----------------------------------------------------------------------------------------
		case IOCTL_POWER_CAPABILITIES: 
        {
            PPOWER_CAPABILITIES ppc;
			RETAILMSG(1, (TEXT("CIS: IOCTL_POWER_CAPABILITIES\r\n")));   
            
			if ( !pdwActualOut || !pBufOut || (dwLenOut < sizeof(POWER_CAPABILITIES)) ) {
                RetVal = FALSE;
                dwErr = ERROR_INVALID_PARAMETER;
                break;
            }
			
            ppc = (PPOWER_CAPABILITIES)pBufOut;
            
            memset(ppc, 0, sizeof(POWER_CAPABILITIES));

            // support D0, D4 
            ppc->DeviceDx = 0x11;

            // Report our power consumption in uAmps rather than mWatts. 
            ppc->Flags = POWER_CAP_PREFIX_MICRO | POWER_CAP_UNIT_AMPS;
            
			// 25 m = 25000 uA
            // TODO: find out a more accurate value
			ppc->Power[D0] = 25000;
            
            *pdwActualOut = sizeof(POWER_CAPABILITIES);
        } break;

		case IOCTL_POWER_SET: 
        {
            CEDEVICE_POWER_STATE NewDx;

			RETAILMSG(1, (TEXT("CIS: IOCTL_POWER_SET\r\n")));   
			
            if ( !pdwActualOut || !pBufOut || (dwLenOut < sizeof(CEDEVICE_POWER_STATE)) ) {
                RetVal = FALSE;
                dwErr = ERROR_INVALID_PARAMETER;
                break;
            }
            
            NewDx = *(PCEDEVICE_POWER_STATE)pBufOut;

            if ( VALID_DX(NewDx) ) {
                switch ( NewDx ) {
                case D0:
					// TODO: ÿλѶ³ʼƺҪ
                    if (pCIS->Dx != D0) {
                        //CIS_PowerUp(hOpenContext);
                        CameraInit();
						EnableInterrupt();

						Camif_Capture(CAPTURE_ON, CAPTURE_ON);

						pCIS->m_dwTimeout = DISPLAY_THREAD_TIMEOUT;
						pCIS->State = RUN;
						
						preview_frame_count = 0;
						codec_frame_count = 0;
                        pCIS->Dx = (_CEDEVICE_POWER_STATE)D0;
                    }
                    break;

                default:
                    if (pCIS->Dx != (_CEDEVICE_POWER_STATE)D4) {
                        CIS_PowerDown(hOpenContext);
                        pCIS->Dx = (_CEDEVICE_POWER_STATE)D4;
                    }
                    break;
                }

                // return our state
                *(PCEDEVICE_POWER_STATE)pBufOut = pCIS->Dx;

                RETAILMSG(1, (TEXT("CIS: IOCTL_POWER_SET: D%u \r\n"), NewDx));

                *pdwActualOut = sizeof(CEDEVICE_POWER_STATE);
            } else {
                RetVal = FALSE;
                dwErr = ERROR_INVALID_PARAMETER;
            }
            
        } break;

        case IOCTL_POWER_GET: 
			RETAILMSG(1, (TEXT("CIS: IOCTL_POWER_GET\r\n")));   
            if ( !pdwActualOut || !pBufOut || (dwLenOut < sizeof(CEDEVICE_POWER_STATE)) ) {
                RetVal = FALSE;
                dwErr = ERROR_INVALID_PARAMETER;
                break;
            }

			*(PCEDEVICE_POWER_STATE)pBufOut = pCIS->Dx;

            RETAILMSG(1, (TEXT("CIS: IOCTL_POWER_GET: D%u \r\n"), pCIS->Dx));

            *pdwActualOut = sizeof(CEDEVICE_POWER_STATE);
	        break;

//-----------------------------------------------------------------------------------------
		case CAM_IOCTL_MOVIE_START:			// for MPEG4
		case IOCTL_CAM_CONT : 
			EnterCriticalSection(&pCIS->CS);
			if (pCIS->State==SUSPEND) {
				CameraInit();
			} else if (pCIS->State == IDLE) {
				CamExitIdle();
			}
			
			EnableInterrupt();

			Camif_Capture(CAPTURE_ON, CAPTURE_ON);

			pCIS->m_dwTimeout = DISPLAY_THREAD_TIMEOUT;
			pCIS->State = RUN;
			LeaveCriticalSection(&pCIS->CS);
			//SetEvent(pCIS->m_hEvent);

			//s2440IOP->GPGCON &= ~(0x3<<24);
			//s2440IOP->GPGCON |= (0x1<<24);		// EINT20
			preview_frame_count = 0;		// for MPEG4
			codec_frame_count = 0;
			running = 1;
			break;
		
		case CAM_IOCTL_MOVIE_STOP:		// for MPEG4
		case IOCTL_CAM_STOP : 
			// Charlie. Stop
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_STOP\r\n")));
			
			DisableInterrupt();
			
			if (s2440INT->INTPND & ( 1 << IRQ_CAM )) s2440INT->INTPND |= ( 1 << IRQ_CAM );
			s2440INT->SRCPND |= ( 1 << IRQ_CAM );
			s2440INT->SUBSRCPND |= (( 1 << IRQ_SUB_CAM_P )|( 1 << IRQ_SUB_CAM_C ));

			/*
			s2440CAM->CICOCTRL |= (1<<2); // CODEC Last IRQ Enable
			s2440CAM->CIPRCTRL |= (1<<2); // PREVIEW Last IRQ Enable
			*/
			Camif_Capture(CAPTURE_OFF, CAPTURE_OFF);
			running = 0;
			break;
			
		case CAM_IOCTL_IMAGE_CHANGE:		// ID=0x521

			if (pBufIn==NULL || dwLenIn<sizeof(CAMINFO))
				RETAILMSG(1,(TEXT("CAM_IOCTL_IMAGE_CHANGE:Parameter Error.\r\n")));

			EnterCriticalSection(&pCIS->CS);
			
			if (pCIS->State==SUSPEND) {
				CameraInit();
				pCIS->State = RUN;
			} else if (pCIS->State == IDLE) {
				CamExitIdle();
				pCIS->State = RUN;
			}

			memcpy((char*)&sCAMINFO, pBufIn, sizeof(sCAMINFO));
						
			RETAILMSG(1,(TEXT("codec image_Xsize:%d, image_YSize:%d, ZoomIndex:%d\r\n"), \
						sCAMINFO.nDestWidth, sCAMINFO.nDestHeight, sCAMINFO.nZoomIndex));			

			DisableInterrupt();

			Camif_Capture(CAPTURE_OFF, CAPTURE_OFF);

			//if((sCAMINFO.nZoomIndex == 2) ||(sCAMINFO.nZoomIndex == 3))
				//CamInit(sCAMINFO.nDestWidth, sCAMINFO.nDestHeight, sCAMINFO.nPreWidth, sCAMINFO.nPreHeight, sCAMINFO.nZoomIndex*80, sCAMINFO.nZoomIndex*60, (U32)(pCIS->m_PhysCodecAddr.LowPart), (U32)(pCIS->m_PhysPreviewAddr.LowPart));
			//else if(sCAMINFO.nZoomIndex == 1)
				CamInit(sCAMINFO.nDestWidth, sCAMINFO.nDestHeight, sCAMINFO.nPreWidth, sCAMINFO.nPreHeight, 0, 0, (U32)(pCIS->m_PhysCodecAddr.LowPart), (U32)(pCIS->m_PhysPreviewAddr.LowPart));				
			//else
				//CamInit(CIF_XSIZE, CIF_YSIZE, sCAMINFO.nPreWidth, sCAMINFO.nPreHeight, 0, 0, (U32)(pCIS->m_PhysCodecAddr.LowPart), (U32)(pCIS->m_PhysPreviewAddr.LowPart));				
				
			LeaveCriticalSection(&pCIS->CS);
			break;

		// for MPEG4
		case CAM_IOCTL_GET_LATEST_FRAME:
		{
			pPixelBlock pb;
			
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:CAM_IOCTL_GET_LATEST_FRAME\r\n")));
			
			if (!pBufIn && !dwLenIn)
				break;
			
			pb = (pPixelBlock)pBufIn;
			// 
			if (pCIS->State==SUSPEND) {
				CameraInit();
				pCIS->State = RUN;
			} else if (pCIS->State == IDLE) {
				CamExitIdle();
				pCIS->State = RUN;
			}

			if (pb->format == 0) {
				if (preview_frame_count < 2) {
					RETAILMSG(1,(TEXT("CAMERA:CAM_IOCTL_GET_LATEST_FRAME - frame not available!!!\r\n")));
					return FALSE;
				}
				if (pBufOut && (dwLenOut>QVGA_XSIZE*QVGA_YSIZE*2))
					CamCopyImage(pBufOut);
				
			} else if (pb->format == 1) {
			
//			Tick_GET_FRAME_PREV = Tick_GET_FRAME_CUR;
//			Tick_GET_FRAME_CUR = GetTickCount();
				if (preview_frame_count < 2) {
					RETAILMSG(1,(TEXT("CAMERA:CAM_IOCTL_GET_LATEST_FRAME - frame not available!!!\r\n")));
					return FALSE;
				}
				if (pBufOut)// && dwLenOut>QVGA_XSIZE*QVGA_YSIZE*2))
					Copy_Cam_Image(pBufOut, VGA_XSIZE, VGA_YSIZE);
				//CamCopyImage(pBufOut);				
			}
		}
			break;
/*
		case IOCTL_CAM_SHOW :
			// Charlie. Show Menu
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SHOW(%x)\r\n"),dwLenIn));
			break;

		case IOCTL_CAM_HIDE :
			// Charlie. Close display window
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_HIDE\r\n")));
			break;

		case IOCTL_CAM_SETPOS : 
			time = GetTickCount();
//			RETAILMSG(MSG_EN_1,(TEXT("Capture time:%d msec\r\n"), (time-old_time)));
			RETAILMSG(MSG_EN_1,(TEXT("Capture time:%d msec\r\n"), (time)));
			old_time = time;
			break;

		case CAM_IOCTL_SAMSUNG_CAM:		// ID=0x520
		{
			//Samsung_camcoder(pBufOut);
			U8 *pImage;
			PINGPONG caminfo;

			RETAILMSG(MSG_EN_1,(_T("CAM_IOCTL_SAMSUNG_CAM\r\n")));
			pImage = pBufOut;

			if( codec_flag )
			{
				caminfo.flag = 1;
				caminfo.y_address = y_address;
				caminfo.cb_address = cb_address;
				caminfo.cr_address = cr_address;
				
				memcpy(pImage, &caminfo, sizeof(PINGPONG));
				codec_flag = 0;
			}
		}	
			break;

		case CAM_IOCTL_SAMSUNG_CAM_PR:	// ID=0x522
		{
			//Samsung_camcoder_pr(pBufOut);
			U8 *pImage;
			PINGPONG_PR prinfo;

			RETAILMSG(MSG_EN_1,(_T("CAM_IOCTL_SAMSUNG_CAM_PR\r\n")));

			pImage = pBufOut;

			if ( rgb_flag )
			{
				prinfo.flag = 1;
				prinfo.rgb_address = rgb_address;
				memcpy(pImage, &prinfo, sizeof(PINGPONG_PR));
				rgb_flag = 0;
			}	
		}	
			break;
		case CAM_IOCTL_SAMSUNG_PREVIEW_START : // ID=0x523
			DRIVER_PREVIEW_ENABLE = 1;
			RETAILMSG(MSG_EN_1,(_T("CAM_IOCTL_SAMSUNG_PREVIEW_START(%x)\r\n"),*pBufOut, DRIVER_PREVIEW_ENABLE));
			break;

		case CAM_IOCTL_SAMSUNG_PREVIEW_STOP : // ID=0x524
			DRIVER_PREVIEW_ENABLE = 2;
			RETAILMSG(MSG_EN_1,(_T("CAM_IOCTL_SAMSUNG_PREVIEW_STOP(%x)\r\n"),*pBufOut, DRIVER_PREVIEW_ENABLE));
			break;
			
		case IOCTL_CAM_SETGAMMA :
			if( *pBufOut == 1 ){
				DRIVER_PREVIEW_ENABLE = 1;
			}
			else if( *pBufOut == 2 ){
				DRIVER_PREVIEW_ENABLE = 2;
			}
			RETAILMSG(MSG_EN_1,(_T("IOCTL_CAM_SETGAMMA:*pBufOut(%x):DRIVER_PREVIEW_ENABLE(%x)\r\n"),*pBufOut, DRIVER_PREVIEW_ENABLE));
			break;
		case CAM_IOCTL_IMAGE_SIZE:		// ID=0x521
		
			RETAILMSG(MSG_EN_1,(TEXT("CAM_IOCTL_IMAGE_SIZE\r\n")));

			if (bIdlePwrDown == TRUE)
			{
				RETAILMSG(PM_MSG, (_T("[CAM_HW] IOControl CAM_START : Turning Codec On\r\n")));

				CamInterface_PowerUp();
	
				bIdlePwrDown = FALSE;
			}

			if( *pBufOut == 1 ){
				image_size = 1;
			}
			else if( *pBufOut == 2 ){
				image_size = 2;
			}
			RETAILMSG(MSG_EN_1,(TEXT("image_size:%d,0x%x,0x%x\r\n"),image_size,pBufOut,*pBufOut));			

			s2440INT->INTMSK |= ( 1 << IRQ_CAM );
			s2440INT->INTSUBMSK |= (( 1 << IRQ_SUB_CAM_P )|( 1 << IRQ_SUB_CAM_C ));

			Camif_Capture(CAPTURE_OFF, CAPTURE_OFF);
			
			if (image_size == 1)
				CamInit(QCIF_XSIZE, QCIF_YSIZE, QCIF_XSIZE, QCIF_YSIZE, 120, 100, COPIFRAMEBUFFER_B, COPIFRAMEBUFFER_A);
			else if (image_size == 2)
				CamInit(CIF_XSIZE, CIF_YSIZE, QCIF_XSIZE, QCIF_YSIZE, 120, 100, COPIFRAMEBUFFER_B, COPIFRAMEBUFFER_A);
			
			break;
*/
		case IOCTL_CAM_COPY :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_COPY(dwLenIn:%x)\r\n"), dwLenIn));
			break;

		case IOCTL_CAM_SETCOLKEY :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETCOLKEY\r\n")));
			break;
			
		case IOCTL_CAM_SETALPHA :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETALPHA\r\n")));
			break;

		case IOCTL_CAM_GETINFO :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_GETINFO\r\n")));
			break;

		case IOCTL_CAM_SETSCALE :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETSCALE\r\n")));
			break;

		case IOCTL_CAM_SETHUE :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETHUE\r\n")));
			break;
#if 0
		case IOCTL_CAM_SETGAMMA :
		// Get which ping -pong buffer have data
			//DisplayEnable = 0;
			Sleep(200);
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_BUFFER:%d\r\n"),buffer_num));
			break;
#endif
		case IOCTL_CAM_SETWBCOORD : 
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETWBCOORD\r\n")));
			break;

		case IOCTL_CAM_SETAELIMIT :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETAELIMIT\r\n")));
			break;

		case IOCTL_CAM_SETADCOFS :	
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETADCOFS\r\n")));
			break;

		case IOCTL_CAM_SETWBGAIN :	
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETWBGAIN\r\n")));
			break;

		case IOCTL_CAM_SETCBCRGAIN :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETCBCRGAIN\r\n")));
			break;

		case IOCTL_CAM_SETLEGAIN :	
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETLEGAIN\r\n")));
			break;

		case IOCTL_CAM_SETBRIGHTNESS :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETBRIGHTNESS\r\n")));
			break;

		case IOCTL_CAM_SETCLIP :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETCLIP\r\n")));
			break;

		case IOCTL_CAM_SETSLICELMT :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETSLICELMT\r\n")));
			break;

		case IOCTL_CAM_WBACCLMT :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_WBACCLMT\r\n")));
			break;

		case IOCTL_CAM_CSEDGELMT :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_CSEDGELMT\r\n")));
			break;

		case IOCTL_CAM_ENVINT :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_ENVINT\r\n")));
			break;

		case IOCTL_CAM_DISVINT :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_DISVINT\r\n")));
			break;

		case IOCTL_CAM_SETCAMERA :
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:IOCTL_CAM_SETCAMERA\r\n")));
			
/*
			RETAILMSG(MSG_EN_1,(_T("---------------------------------------\r\n")));
			RETAILMSG(MSG_EN_1,(_T("ImgCptEn_PRSC = %d\r\n"), (s2440CAM->CIPRSTATUS>>21)&0x1));
			RETAILMSG(MSG_EN_1,(_T("PrScalerStart = %d\r\n"), (s2440CAM->CIPRSCCTRL>>15)&0x1));
			RETAILMSG(MSG_EN_1,(_T("---------------------------------------\r\n")));
			RETAILMSG(MSG_EN_1,(_T("CoScalerStart  = %d\r\n"), (s2440CAM->CICOSCCTRL>>15)&0x1));
			RETAILMSG(MSG_EN_1,(_T("VSYNC          = %d\r\n"), (s2440CAM->CICOSTATUS>>28)&0x1));
			RETAILMSG(MSG_EN_1,(_T("ImgCptEn_Camif = %d\r\n"), (s2440CAM->CICOSTATUS>>22)&0x1));
			RETAILMSG(MSG_EN_1,(_T("ImgCptEn_CoSC  = %d\r\n"), (s2440CAM->CICOSTATUS>>21)&0x1));
			RETAILMSG(MSG_EN_1,(_T("VSYNC_A  = %d\r\n"), (s2440CAM->CICOSTATUS>>20)&0x1));
			RETAILMSG(MSG_EN_1,(_T("VSYNC_B  = %d\r\n"), (s2440CAM->CICOSTATUS>>19)&0x1));
			RETAILMSG(MSG_EN_1,(_T("---------------------------------------\r\n")));
			RETAILMSG(MSG_EN_1,(_T("ImgCptEn      = %d\r\n"), (s2440CAM->CIIMGCPT>>31)&0x1));
			RETAILMSG(MSG_EN_1,(_T("ImgCptEn_CoSc = %d\r\n"), (s2440CAM->CIIMGCPT>>30)&0x1));
			RETAILMSG(MSG_EN_1,(_T("ImgCptEn_PrSc = %d\r\n"), (s2440CAM->CIIMGCPT>>29)&0x1));
			RETAILMSG(MSG_EN_1,(_T("----------------- %d ----------------------\r\n"), mInitialized));
			SetEvent(pCIS->m_hEvent);
*/
			break;

		default : 
			RETAILMSG(MSG_EN_1,(TEXT("CAMERA:Ioctl code = 0x%x\r\n"), dwCode));
			return FALSE;

	}
	return TRUE;
} 

